@@ -162,6 +162,13 @@ class ArgumentParser(argparse.ArgumentParser):
         return [arg_line.strip()]
 
 
+class AppendCommaSeparatedArg(argparse.Action):
+    def __call__(self, parser, args, value, option_string=None):
+        result = getattr(args, self.dest, [])
+        result.extend(value.split(','))
+        setattr(args, self.dest, result)
+
+
 def parse_args(arglist=None):
     '''Parse autopkgtest command line arguments.
 
@@ -299,6 +306,11 @@ for details.'''
     g_setup.add_argument('--env', metavar='VAR=value',
                          action='append', default=[],
                          help='Set arbitrary environment variable for builds and test')
+    g_setup.add_argument('--ignore-restrictions', default=[],
+                         metavar='RESTRICTION[,RESTRICTION...]',
+                         action=AppendCommaSeparatedArg,
+                         help='Run tests even if these restrictions would '
+                         'normally prevent it')
 
     # privileges
     g_priv = parser.add_argument_group('user/privilege handling options')
@@ -354,6 +366,9 @@ for details.'''
         'they will be treated as flaky tests. With "skip" these tests will be '
         'skipped immediately and will not be run. With "run" the restriction '
         'is basically ignored.')
+    g_misc.add_argument(
+        '-V', '--validate', action='store_true', default=False,
+        help='validate the test control file and exit')
     g_misc.add_argument(
         '-h', '--help', action='help', default=argparse.SUPPRESS,
         help='show this help message and exit')
