@@ -99,9 +99,6 @@ class Test:
         '''
         if '/' in name:
             raise Unsupported(name, 'test name may not contain / character')
-        for r in restrictions:
-            if r not in known_restrictions:
-                raise Unsupported(name, 'unknown restriction %s' % r)
 
         if not ((path is None) ^ (command is None)):
             raise InvalidControl(name, 'Test must have either path or command')
@@ -151,37 +148,43 @@ class Test:
         else:
             adtlog.report(self.name, 'FAIL ' + reason)
 
-    def check_testbed_compat(self, caps):
+    def check_testbed_compat(self, caps, ignore_restrictions=()):
         '''Check for restrictions incompatible with test bed capabilities.
 
         Raise Unsupported exception if there are any.
         '''
-        if 'isolation-container' in self.restrictions and \
+        effective = set(self.restrictions) - set(ignore_restrictions)
+
+        for r in effective:
+            if r not in known_restrictions:
+                raise Unsupported(self.name, 'unknown restriction %s' % r)
+
+        if 'isolation-container' in effective and \
            'isolation-container' not in caps and \
            'isolation-machine' not in caps:
             raise Unsupported(self.name,
                               'Test requires container-level isolation but '
                               'testbed does not provide that')
 
-        if 'isolation-machine' in self.restrictions and \
+        if 'isolation-machine' in effective and \
            'isolation-machine' not in caps:
             raise Unsupported(self.name,
                               'Test requires machine-level isolation but '
                               'testbed does not provide that')
 
-        if 'breaks-testbed' in self.restrictions and \
+        if 'breaks-testbed' in effective and \
            'revert-full-system' not in caps:
             raise Unsupported(self.name,
                               'Test breaks testbed but testbed does not '
                               'provide revert-full-system')
 
-        if 'needs-root' in self.restrictions and \
+        if 'needs-root' in effective and \
            'root-on-testbed' not in caps:
             raise Unsupported(self.name,
                               'Test needs root on testbed which is not '
                               'available')
 
-        if 'needs-reboot' in self.restrictions and \
+        if 'needs-reboot' in effective and \
            'reboot' not in caps:
             raise Unsupported(self.name,
                               'Test needs to reboot testbed but testbed does '
@@ -240,7 +243,7 @@ def parse_rfc822(path):
 def _debian_check_unknown_fields(name, record):
     unknown_keys = set(record.keys()).difference(
         {'Tests', 'Test-command', 'Restrictions', 'Features',
-         'Depends', 'Tests-directory', 'Classes'})
+         'Depends', 'Tests-directory', 'Classes', 'Architecture'})
     if unknown_keys:
         raise Unsupported(name, 'unknown field %s' % unknown_keys.pop())
 
@@ -434,10 +437,73 @@ def _autodep8(srcdir):
     return None
 
 
+def _matches_architecture(host_arch, arch_wildcard):
+    try:
+        subprocess.check_call(['perl', '-mDpkg::Arch', '-e',
+                               'exit(!Dpkg::Arch::debarch_is(shift, shift))',
+                               host_arch, arch_wildcard])
+    except subprocess.CalledProcessError as e:
+        # returns 1 if host_arch is not matching arch_wildcard; other
+        # errors shouldn't be ignored
+        if e.returncode != 1:
+            raise
+        return False
+    return True
+
+
+def _check_architecture(name, testbed_arch, architectures):
+    '''Check if testbed_arch is supported by the architectures
+
+    The architecture list comes in two variants, positive: only this
+    arch is supported (arch may be a wildcard) and negative: this arch
+    is not supported (arch may be a wildcard). If there is any
+    positive arch, every arch not explicitly listed is skipped. Debian
+    Policy 7.1 explains that for (Build-)Depends it's not allowed to
+    mix positive and negative, so let's not do either. The list can
+    also be empty. Empty and ["any"] are the same, "all" isn't
+    allowed.
+    '''
+
+    if "all" in architectures:
+        raise Unsupported(name, "Arch 'all' not allowed in Architecture field")
+
+    if len(architectures) == 0 or architectures == ["any"]:
+        return
+
+    any_negative = False
+    any_positive = False
+    for arch in architectures:
+        if arch[0] == "!":
+            any_negative = True
+            if _matches_architecture(testbed_arch, arch[1:]):
+                raise Unsupported(name, "Test declares architecture as not " +
+                                  "supported: %s" % arch)
+        if arch[0] != "!":
+            any_positive = True
+
+    if any_positive:
+        if any_negative:
+            raise Unsupported(name, "It is not permitted for some archs to " +
+                              "be prepended by an exclamation mark while " +
+                              "others aren't")
+        arch_matched = False
+        for arch in architectures:
+            if _matches_architecture(testbed_arch, arch):
+                arch_matched = True
+
+        if not arch_matched:
+            raise Unsupported(name, "Test lists explicitly supported " +
+                              "architectures, but the current architecture " +
+                              "%s isn't listed." % arch)
+
+
 def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
-                        auto_control=True):
+                        auto_control=True, ignore_restrictions=(), testname=None):
     '''Parse test descriptions from a Debian DEP-8 source dir
 
+    @ignore_restrictions: If we would skip the test due to these restrictions,
+                          run it anyway
+
     You can specify an alternative path for the control file (default:
     srcdir/debian/tests/control).
 
@@ -502,9 +568,13 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
                         '*', 'only one test-name feature allowed')
                 feature_test_name = details[1]
                 features.append(feature)
+            architectures = record.get('Architecture', '').replace(
+                ',', ' ').split()
 
             if 'Tests' in record:
                 test_names = record['Tests'].replace(',', ' ').split()
+                if len(test_names) == 0:
+                    raise InvalidControl('*', '"Tests" field is empty')
                 (depends, synth_depends) = _parse_debian_depends(
                     test_names[0],
                     record.get('Depends', '@'),
@@ -521,12 +591,15 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
                 for n in test_names:
                     try:
                         _debian_check_unknown_fields(n, record)
+                        _check_architecture(n, testbed_arch, architectures)
+
                         test = Test(n, os.path.join(test_dir, n), None,
                                     restrictions, features, depends, [], [], synth_depends)
-                        test.check_testbed_compat(testbed_caps)
+                        test.check_testbed_compat(testbed_caps, ignore_restrictions)
                     except Unsupported as u:
-                        u.report()
-                        some_skipped = True
+                        if testname is None or n == testname:
+                            u.report()
+                            some_skipped = True
                     else:
                         tests.append(test)
             elif 'Test-command' in record:
@@ -542,16 +615,18 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
                 else:
                     name = feature_test_name
                 _debian_check_unknown_fields(name, record)
+                _check_architecture(name, testbed_arch, architectures)
                 test = Test(name, None, command, restrictions, features,
                             depends, [], [], synth_depends)
-                test.check_testbed_compat(testbed_caps)
+                test.check_testbed_compat(testbed_caps, ignore_restrictions)
                 tests.append(test)
             else:
                 raise InvalidControl('*', 'missing "Tests" or "Test-Command"'
                                      ' field')
         except Unsupported as u:
-            u.report()
-            some_skipped = True
+            if testname is None or n == testname:
+                u.report()
+                some_skipped = True
 
     return (tests, some_skipped)
 
@@ -561,7 +636,7 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
 #
 
 def parse_click_manifest(manifest, testbed_caps, clickdeps, use_installed,
-                         srcdir=None):
+                         srcdir=None, ignore_restrictions=(), testname=None):
     '''Parse test descriptions from a click manifest.
 
     @manifest: String with the click manifest
@@ -569,6 +644,9 @@ def parse_click_manifest(manifest, testbed_caps, clickdeps, use_installed,
     @clickdeps: paths of click packages that these tests need
     @use_installed: True if test expects the described click to be installed
                     already
+    @ignore_restrictions: If we would skip the test due to these restrictions,
+                          run it anyway
+    @testname: If we're asked a specific test, don't report on others
 
     Return (source_dir, list of Test objects, some_skipped). If this encounters
     any invalid restrictions, fields, or test restrictions which cannot be met
@@ -634,11 +712,12 @@ def parse_click_manifest(manifest, testbed_caps, clickdeps, use_installed,
             test = Test(name, desc.get('path'), desc.get('command'),
                         desc.get('restrictions', []), desc.get('features', []),
                         desc.get('depends', []), clickdeps, installed_clicks, [])
-            test.check_testbed_compat(testbed_caps)
+            test.check_testbed_compat(testbed_caps, ignore_restrictions)
             tests.append(test)
         except Unsupported as u:
-            u.report()
-            some_skipped = True
+            if testname is None or name == testname:
+                u.report()
+                some_skipped = True
 
     if srcdir is None:
         # do we have an x-source/vcs-bzr link?
@@ -665,7 +744,7 @@ def parse_click_manifest(manifest, testbed_caps, clickdeps, use_installed,
     return (srcdir, tests, some_skipped)
 
 
-def parse_click(clickpath, testbed_caps, srcdir=None):
+def parse_click(clickpath, testbed_caps, srcdir=None, testname=None):
     '''Parse test descriptions from a click package.
 
     Return (source_dir, list of Test objects, some_skipped). If this encounters
@@ -687,4 +766,4 @@ def parse_click(clickpath, testbed_caps, srcdir=None):
         pkg.close()
 
     return parse_click_manifest(manifest, testbed_caps, [clickpath], False,
-                                srcdir)
+                                srcdir, testname=testname)
